@@ -568,6 +568,8 @@ Gsub::Gsub(const Data &d, ErrorHandler *errh) throw (Error)
     // Offset	ScriptList
     // Offset	FeatureList
     // Offset	LookupList
+    if (d.length() == 0)
+	throw BlankTable("GSUB");
     if (d.u16(0) != 1)
 	throw Format("GSUB");
     if (_script_list.assign(d.offset_subtable(4), errh) < 0)
