@@ -25,6 +25,7 @@
 #include <lcdf/hashmap.hh>
 #include <errno.h>
 #include <ctype.h>
+#include <algorithm>
 #ifndef __KERNEL__
 # include <stdlib.h>
 #endif
@@ -93,6 +94,17 @@ ErrorHandler::fatal(const char *format, ...)
   return ERROR_RESULT;
 }
 
+int
+ErrorHandler::fatal(int exit_status, const char *format, ...)
+{
+  assert(exit_status >= 0 && exit_status < 0x1000);
+  va_list val;
+  va_start(val, format);
+  verror((Seriousness) (ERR_MIN_FATAL + (exit_status << ERRVERBOSITY_SHIFT)), String(), format, val);
+  va_end(val);
+  return ERROR_RESULT;
+}
+
 void
 ErrorHandler::ldebug(const String &where, const char *format, ...)
 {
@@ -141,6 +153,17 @@ ErrorHandler::lfatal(const String &where, const char *format, ...)
   return ERROR_RESULT;
 }
 
+int
+ErrorHandler::lfatal(int exit_status, const String &where, const char *format, ...)
+{
+  assert(exit_status >= 0 && exit_status < 0x1000);
+  va_list val;
+  va_start(val, format);
+  verror((Seriousness) (ERR_MIN_FATAL + (exit_status << ERRVERBOSITY_SHIFT)), where, format, val);
+  va_end(val);
+  return ERROR_RESULT;
+}
+
 String
 ErrorHandler::make_text(Seriousness seriousness, const char *format, ...)
 {
@@ -211,20 +234,13 @@ do_number_flags(char *pos, char *after_last, int base, int flags,
 }
 
 String
-ErrorHandler::make_text(Seriousness seriousness, const char *s, va_list val)
+ErrorHandler::make_text(Seriousness, const char *s, va_list val)
 {
   StringAccum msg;
   char numbuf[NUMBUF_SIZE];	// for numerics
   String placeholder;		// to ensure temporaries aren't destroyed
-  String s_placeholder;		// ditto, in case we change `s'
   numbuf[NUMBUF_SIZE-1] = 0;
   
-  if (seriousness >= ERR_MIN_WARNING && seriousness < ERR_MIN_ERROR) {
-    // prepend `warning: ' to every line
-    s_placeholder = prepend_lines("warning: ", s);
-    s = s_placeholder.c_str();
-  }
-  
   // declare and initialize these here to make gcc shut up about possible 
   // use before initialization
   int flags = 0;
@@ -475,39 +491,35 @@ ErrorHandler::make_text(Seriousness seriousness, const char *s, va_list val)
 }
 
 String
-ErrorHandler::decorate_text(Seriousness, const String &prefix, const String &landmark, const String &text)
+ErrorHandler::decorate_text(Seriousness seriousness, const String &landmark, const String &text)
 {
   String new_text;
   
-  if (!landmark)
+  // prepend 'warning: ' to every line if appropriate
+  if (seriousness >= ERR_MIN_WARNING && seriousness < ERR_MIN_ERROR)
+    new_text = prepend_lines("warning: ", text);
+  else
     new_text = text;
 
-  else if (landmark.length() > 2 && landmark[0] == '\\'
-	   && landmark.back() == '\\')
+  // handle landmark
+  if (landmark
+      && !(landmark.length() > 2 && landmark[0] == '\\'
+	   && landmark.back() == '\\')) {
     // ignore special-purpose landmarks (begin and end with a backslash `\')
-    new_text = text;
-
-  else {
     // fix landmark: skip trailing spaces and trailing colon
     int i, len = landmark.length();
     for (i = len - 1; i >= 0; i--)
       if (!isspace(landmark[i]))
 	break;
+    if (i >= 0 && landmark[i] == ':')
+      i--;
 
     // prepend landmark, unless all spaces
-    if (i >= 0) {
-      if (landmark[i] == ':')
-	i--;
-      new_text = prepend_lines(landmark.substring(0, i+1) + ": ", text);
-    } else
-      new_text = text;
+    if (i >= 0)
+      new_text = prepend_lines(landmark.substring(0, i+1) + ": ", new_text);
   }
 
-  // prepend prefix, if any
-  if (prefix)
-    return prepend_lines(prefix, new_text);
-  else
-    return new_text;
+  return new_text;
 }
 
 int
@@ -515,7 +527,7 @@ ErrorHandler::verror(Seriousness seriousness, const String &where,
 		     const char *s, va_list val)
 {
   String text = make_text(seriousness, s, val);
-  text = decorate_text(seriousness, String(), where, text);
+  text = decorate_text(seriousness, where, text);
   handle_text(seriousness, text);
   return count_error(seriousness, text);
 }
@@ -525,7 +537,7 @@ ErrorHandler::verror_text(Seriousness seriousness, const String &where,
 			  const String &text)
 {
   // text is already made
-  String dec_text = decorate_text(seriousness, String(), where, text);
+  String dec_text = decorate_text(seriousness, where, text);
   handle_text(seriousness, dec_text);
   return count_error(seriousness, dec_text);
 }
@@ -542,13 +554,15 @@ ErrorHandler::prepend_lines(const String &prepend, const String &text)
     return text;
   
   StringAccum sa;
-  int pos = 0, nl;
-  while ((nl = text.find_left('\n', pos)) >= 0) {
-    sa << prepend << text.substring(pos, nl - pos + 1);
-    pos = nl + 1;
+  const char *begin = text.begin();
+  const char *end = text.end();
+  const char *nl;
+  while ((nl = std::find(begin, end, '\n')) < end) {
+    sa << prepend << text.substring(begin, nl + 1);
+    begin = nl + 1;
   }
-  if (pos < text.length())
-    sa << prepend << text.substring(pos);
+  if (begin < end)
+    sa << prepend << text.substring(begin, end);
   
   return sa.take_string();
 }
@@ -584,19 +598,15 @@ FileErrorHandler::FileErrorHandler(FILE *f, const String &context)
 void
 FileErrorHandler::handle_text(Seriousness seriousness, const String &message)
 {
-  int pos = 0, nl;
-  while ((nl = message.find_left('\n', pos)) >= 0) {
-    String s = _context + message.substring(pos, nl - pos) + "\n";
-    fwrite(s.data(), 1, s.length(), _f);
-    pos = nl + 1;
-  }
-  if (pos < message.length()) {
-    String s = _context + message.substring(pos) + "\n";
-    fwrite(s.data(), 1, s.length(), _f);
-  }
+  String text = prepend_lines(_context, message);
+  if (text && text.back() != '\n')
+    text += '\n';
+  fwrite(text.data(), 1, text.length(), _f);
   
-  if (seriousness >= ERR_MIN_FATAL)
-    exit(1);
+  if (seriousness >= ERR_MIN_FATAL) {
+    int exit_status = (seriousness - ERR_MIN_FATAL) >> ERRVERBOSITY_SHIFT;
+    exit(exit_status);
+  }
 }
 #endif
 
@@ -653,7 +663,6 @@ ErrorHandler *
 ErrorHandler::static_initialize(ErrorHandler *default_handler)
 {
   the_default_handler = default_handler;
-  the_silent_handler = new SilentErrorHandler;
   return the_default_handler;
 }
 
@@ -686,7 +695,8 @@ ErrorHandler::default_handler()
 ErrorHandler *
 ErrorHandler::silent_handler()
 {
-  assert(the_silent_handler != 0);
+  if (!the_silent_handler)
+    the_silent_handler = new SilentErrorHandler;
   return the_silent_handler;
 }
 
@@ -726,9 +736,9 @@ ErrorVeneer::make_text(Seriousness seriousness, const char *s, va_list val)
 }
 
 String
-ErrorVeneer::decorate_text(Seriousness seriousness, const String &prefix, const String &landmark, const String &text)
+ErrorVeneer::decorate_text(Seriousness seriousness, const String &landmark, const String &text)
 {
-  return _errh->decorate_text(seriousness, prefix, landmark, text);
+  return _errh->decorate_text(seriousness, landmark, text);
 }
 
 void
@@ -750,13 +760,15 @@ ErrorVeneer::count_error(Seriousness seriousness, const String &text)
 
 ContextErrorHandler::ContextErrorHandler(ErrorHandler *errh,
 					 const String &context,
-					 const String &indent)
-  : ErrorVeneer(errh), _context(context), _indent(indent)
+					 const String &indent,
+					 const String &context_landmark)
+  : ErrorVeneer(errh), _context(context), _indent(indent),
+    _context_landmark(context_landmark)
 {
 }
 
 String
-ContextErrorHandler::decorate_text(Seriousness seriousness, const String &prefix, const String &landmark, const String &text)
+ContextErrorHandler::decorate_text(Seriousness seriousness, const String &landmark, const String &text)
 {
   String context_lines;
   if (_context) {
@@ -765,13 +777,13 @@ ContextErrorHandler::decorate_text(Seriousness seriousness, const String &prefix
     if (_errh->min_verbosity() > ERRVERBOSITY_CONTEXT)
       _context = _indent = String();
     else {
-      context_lines = _errh->decorate_text(ERR_MESSAGE, String(), landmark, _context);
+      context_lines = _errh->decorate_text(ERR_MESSAGE, (_context_landmark ? _context_landmark : landmark), _context);
       if (context_lines && context_lines.back() != '\n')
 	context_lines += '\n';
       _context = String();
     }
   }
-  return context_lines + _errh->decorate_text(seriousness, String(), landmark, prepend_lines(_indent + prefix, text));
+  return context_lines + _errh->decorate_text(seriousness, (landmark ? landmark : _context_landmark), prepend_lines(_indent, text));
 }
 
 
@@ -786,26 +798,9 @@ PrefixErrorHandler::PrefixErrorHandler(ErrorHandler *errh,
 }
 
 String
-PrefixErrorHandler::decorate_text(Seriousness seriousness, const String &prefix, const String &landmark, const String &text)
+PrefixErrorHandler::decorate_text(Seriousness seriousness, const String &landmark, const String &text)
 {
-  return _errh->decorate_text(seriousness, _prefix + prefix, landmark, text);
-}
-
-
-//
-// INDENT ERROR HANDLER
-//
-
-IndentErrorHandler::IndentErrorHandler(ErrorHandler *errh,
-				       const String &indent)
-  : ErrorVeneer(errh), _indent(indent)
-{
-}
-
-String
-IndentErrorHandler::decorate_text(Seriousness seriousness, const String &prefix, const String &landmark, const String &text)
-{
-  return _errh->decorate_text(seriousness, String(), landmark, prepend_lines(_indent + prefix, text));
+  return _errh->decorate_text(seriousness, landmark, prepend_lines(_prefix, text));
 }
 
 
@@ -819,12 +814,12 @@ LandmarkErrorHandler::LandmarkErrorHandler(ErrorHandler *errh, const String &lan
 }
 
 String
-LandmarkErrorHandler::decorate_text(Seriousness seriousness, const String &prefix, const String &lm, const String &text)
+LandmarkErrorHandler::decorate_text(Seriousness seriousness, const String &lm, const String &text)
 {
   if (lm)
-    return _errh->decorate_text(seriousness, prefix, lm, text);
+    return _errh->decorate_text(seriousness, lm, text);
   else
-    return _errh->decorate_text(seriousness, prefix, _landmark, text);
+    return _errh->decorate_text(seriousness, _landmark, text);
 }
 
 
