@@ -31,6 +31,8 @@ static HashMap<String, int> glyphlist(-1);
 static PermString::Initializer perm_initializer;
 PermString DvipsEncoding::dot_notdef(".notdef");
 
+#define NEXT_GLYPH_NAME(gn)	("/" + (gn))
+
 int
 DvipsEncoding::parse_glyphlist(String text)
 {
@@ -71,7 +73,8 @@ DvipsEncoding::parse_glyphlist(String text)
 		    next++;
 	    pos = next - data;
 	    if (*next == ';') {	// read another possibility
-		glyph_name += "/"; // XXX handles "DDDD;DDDD DDDD;DDDD" badly
+		glyph_name = NEXT_GLYPH_NAME(glyph_name);
+		// XXX given "DDDD;EEEE EEEE;FFFF", will not store "FFFF"
 		goto read_uni;
 	    }
 	} else
@@ -80,22 +83,6 @@ DvipsEncoding::parse_glyphlist(String text)
     }
 }
 
-static int
-uniparsenumber(const char *a, int len)
-{
-    int value = 0;
-    for (int i = 0; i < len; a++, i++)
-	if (*a >= '0' && *a <= '9')
-	    value = (value << 4) | (*a - '0');
-	else if (*a >= 'A' && *a <= 'F')
-	    value = (value << 4) | (*a - 'A' + 10);
-	else
-	    return -1;
-    return ((value >= 0 && value <= 0xD7FF)
-	    || (value >= 0xE000 && value <= 0x10FFFF)
-	    ? value : -1);
-}
-
 void
 DvipsEncoding::glyphname_unicode(String gn, Vector<int> &unis, bool *more)
 {    
@@ -115,6 +102,7 @@ DvipsEncoding::glyphname_unicode(String gn, Vector<int> &unis, bool *more)
 
 	// check glyphlist
 	int value = glyphlist[component];
+	uint32_t uval;
 	if (value >= 0) {
 	    unis.push_back(value & ~GLYPHLIST_MORE);
 	    if (more && (value & GLYPHLIST_MORE) && !gn)
@@ -123,9 +111,9 @@ DvipsEncoding::glyphname_unicode(String gn, Vector<int> &unis, bool *more)
 		   && (component.length() % 4) == 3
 		   && memcmp(component.data(), "uni", 3) == 0) {
 	    int old_size = unis.size();
-	    for (int i = 3; i < component.length(); i += 4)
-		if ((value = uniparsenumber(component.data() + i, 4)) >= 0)
-		    unis.push_back(value);
+	    for (const char* s = component.begin() + 3; s < component.end(); s += 4)
+		if (parse_unicode_number(s, s + 4, -1, uval))
+		    unis.push_back(uval);
 		else {
 		    unis.resize(old_size);
 		    break;
@@ -133,8 +121,8 @@ DvipsEncoding::glyphname_unicode(String gn, Vector<int> &unis, bool *more)
 	} else if (component.length() >= 5
 		   && component.length() <= 7
 		   && component[0] == 'u'
-		   && (value = uniparsenumber(component.data() + 1, component.length() - 1)) >= 0)
-	    unis.push_back(value);
+		   && parse_unicode_number(component.begin() + 1, component.end(), -1, uval))
+	    unis.push_back(uval);
     }
 }
 
@@ -158,7 +146,6 @@ DvipsEncoding::encode(int e, PermString what)
     if (e >= _e.size())
 	_e.resize(e + 1, dot_notdef);
     _e[e] = what;
-    _unicodes.clear();		// _unicodes isn't good any more
 }
 
 int
@@ -307,6 +294,7 @@ find_ligkern_op(const String &s)
 int
 DvipsEncoding::parse_ligkern_words(Vector<String> &v, ErrorHandler *errh)
 {
+    _file_had_ligkern = true;
     int op;
     if (v.size() == 3) {
 	if (v[0] == "||" && v[1] == "=") {
@@ -378,7 +366,7 @@ DvipsEncoding::parse_unicoding_words(Vector<String> &v, ErrorHandler *errh)
 	    } else {
 		_unicoding.push_back(uni);
 		while (more) {
-		    v[i] += "/";
+		    v[i] = NEXT_GLYPH_NAME(v[i]);
 		    if ((uni = glyphname_unicode(v[i], &more)) >= 0)
 			_unicoding.push_back(uni);
 		}
@@ -393,7 +381,6 @@ DvipsEncoding::parse_unicoding_words(Vector<String> &v, ErrorHandler *errh)
 int
 DvipsEncoding::parse_words(const String &s, int (DvipsEncoding::*method)(Vector<String> &, ErrorHandler *), ErrorHandler *errh)
 {
-    _file_had_comments = true;
     Vector<String> words;
     const char *data = s.data();
     const char *end = s.end();
@@ -424,14 +411,14 @@ landmark(const String &filename, int line)
 }
 
 int
-DvipsEncoding::parse(String filename, ErrorHandler *errh)
+DvipsEncoding::parse(String filename, bool ignore_ligkern, bool ignore_other, ErrorHandler *errh)
 {
     int before = errh->nerrors();
     String s = read_file(filename, errh);
     if (errh->nerrors() != before)
 	return -1;
     _filename = filename;
-    _file_had_comments = false;
+    _file_had_ligkern = false;
     filename = printable_filename(filename);
     int pos = 0, line = 1;
 
@@ -457,32 +444,35 @@ DvipsEncoding::parse(String filename, ErrorHandler *errh)
     while ((token = comment_tokenize(s, pos, line)))
 	if (token.length() >= 8
 	    && memcmp(token.data(), "LIGKERN", 7) == 0
-	    && isspace(token[7])) {
+	    && isspace(token[7])
+	    && !ignore_ligkern) {
 	    lerrh.set_landmark(landmark(filename, line));
 	    parse_words(token.substring(8), &DvipsEncoding::parse_ligkern_words, &lerrh);
 	    
 	} else if (token.length() >= 9
 		   && memcmp(token.data(), "LIGKERNX", 8) == 0
-		   && isspace(token[8])) {
+		   && isspace(token[8])
+		   && !ignore_ligkern) {
 	    lerrh.set_landmark(landmark(filename, line));
 	    parse_words(token.substring(9), &DvipsEncoding::parse_ligkern_words, &lerrh);
 	    
 	} else if (token.length() >= 10
 		   && memcmp(token.data(), "UNICODING", 9) == 0
-		   && isspace(token[9])) {
+		   && isspace(token[9])
+		   && !ignore_other) {
 	    lerrh.set_landmark(landmark(filename, line));
 	    parse_words(token.substring(10), &DvipsEncoding::parse_unicoding_words, &lerrh);
 	    
 	} else if (token.length() >= 13
 		   && memcmp(token.data(), "CODINGSCHEME", 12) == 0
-		   && isspace(token[12])) {
+		   && isspace(token[12])
+		   && !ignore_other) {
 	    int p = 13;
 	    while (p < token.length() && isspace(token[p]))
 		p++;
 	    int pp = token.length() - 1;
 	    while (pp > p && isspace(token[pp]))
 		pp--;
-	    _file_had_comments = true;
 	    _coding_scheme = token.substring(p, pp - p);
 	    if (_coding_scheme.length() > 39)
 		lerrh.lwarning(landmark(filename, line), "only first 39 chars of CODINGSCHEME are significant");
@@ -527,14 +517,113 @@ map_uni(uint32_t uni, const Efont::OpenType::Cmap &cmap, const Metrics &m)
 	return cmap.map_uni(uni);
 }
 
+bool
+DvipsEncoding::x_unicodes(PermString chname, Vector<uint32_t> &unicodes) const
+{
+    int i = _unicoding_map[chname];
+    if (i >= 0) {
+	for (; _unicoding[i] >= 0; i++)
+	    unicodes.push_back(_unicoding[i]);
+	return true;
+    } else {
+	bool more;
+	if ((i = glyphname_unicode(chname, &more)) >= 0)
+	    unicodes.push_back(i);
+	if (more) {		// might be multiple possibilities
+	    String gn = chname;
+	    do {
+		gn = NEXT_GLYPH_NAME(gn);
+		if ((i = glyphname_unicode(gn, &more)) >= 0)
+		    unicodes.push_back(i);
+	    } while (more);
+	}
+	return false;
+    }
+}
+
+	
+void
+DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap, Efont::Cff::Font *font, Secondary *secondary, bool literal, ErrorHandler *errh)
+{
+    // first pass: without secondaries
+    for (int code = 0; code < _e.size(); code++) {
+	PermString chname = _e[code];
+
+	// the altselector character has its own glyph name
+	if (code == _altselector_char && !literal)
+	    chname = "altselector";
+
+	// common case: skip .notdef
+	if (chname == dot_notdef)
+	    continue;
+
+	// find all Unicodes
+	Vector<uint32_t> unicodes;
+	bool unicodes_explicit = x_unicodes(chname, unicodes);
+
+	// find first Unicode supported by the font
+	Efont::OpenType::Glyph glyph = 0;
+	uint32_t glyph_uni = (unicodes.size() ? unicodes[0] : 0);
+	for (uint32_t *u = unicodes.begin(); u < unicodes.end() && !glyph; u++)
+	    if ((glyph = map_uni(*u, cmap, metrics)) > 0)
+		glyph_uni = *u;
+
+	// do not use a Unicode-mapped glyph if literal
+	if (literal)
+	    glyph = 0;
+	
+	// use named glyph if (1) name contains a dot, or (2) nothing found
+	if (font
+	    && !unicodes_explicit
+	    && (glyph <= 0
+		|| std::find(chname.begin(), chname.end(), '.') < chname.end()))
+	    if (Efont::OpenType::Glyph named_glyph = font->glyphid(chname))
+		glyph = named_glyph;
+
+	// assign slot
+	if (glyph >= 0)
+	    metrics.encode(code, glyph_uni, glyph);
+    }
+
+    // second pass: with secondaries
+    for (int code = 0; secondary && code < _e.size(); code++)
+	if (_e[code] != dot_notdef && metrics.glyph(code) <= 0) {
+	    PermString chname = _e[code];
+
+	    // the altselector character has its own glyph name
+	    if (code == _altselector_char)
+		chname = "altselector";
+	
+	    // loop over all Unicodes
+	    Vector<uint32_t> unicodes;
+	    (void) x_unicodes(chname, unicodes);
+	    for (uint32_t *u = unicodes.begin(); u < unicodes.end(); u++)
+		if (secondary->encode_uni(code, chname, *u, metrics, errh))
+		    break;
+	}
+
+    // final pass: complain
+    for (int code = 0; code < _e.size(); code++)
+	if (_e[code] != dot_notdef && metrics.glyph(code) <= 0)
+	    bad_codepoint(code);
+
+    metrics.set_coding_scheme(_coding_scheme);
+}
+
+
+#if 0
 void
 DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap, Efont::Cff::Font *font, Secondary *secondary, ErrorHandler *errh)
 {
     for (int i = 0; i < _e.size(); i++) {
 	PermString chname = _e[i];
+
+	// the altselector character has its own glyph name
 	if (i == _altselector_char)
 	    chname = "altselector";
-	else if (chname == dot_notdef)
+
+	// common case: skip .notdef
+	if (chname == dot_notdef)
 	    continue;
 	
 	Efont::OpenType::Glyph gid = 0;
@@ -590,7 +679,7 @@ DvipsEncoding::make_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap,
 }
 
 void
-DvipsEncoding::make_literal_metrics(Metrics &metrics, Efont::Cff::Font *font)
+DvipsEncoding::make_literal_metrics(Metrics &metrics, const Efont::OpenType::Cmap &cmap, Efont::Cff::Font *font)
 {
     for (int i = 0; i < _e.size(); i++)
 	if (_e[i] != dot_notdef) {
@@ -620,16 +709,7 @@ DvipsEncoding::unicodes() const
     }
     return _unicodes;
 }
-
-int
-DvipsEncoding::encoding_of_unicode(uint32_t uni) const
-{
-    (void) unicodes();		// make _unicodes array
-    for (int i = 0; i < _unicodes.size(); i++)
-	if (_unicodes[i] == uni)
-	    return i;
-    return -1;
-}
+#endif
 
 void
 DvipsEncoding::apply_ligkern_lig(Metrics &metrics, ErrorHandler *errh) const
