@@ -21,6 +21,7 @@
 #include <efont/t1bounds.hh>
 #include <efont/otfcmap.hh>
 #include <efont/otfgsub.hh>
+#include "glyphfilter.hh"
 #include "metrics.hh"
 #include "dvipsencoding.hh"
 #include "automatic.hh"
@@ -79,22 +80,26 @@ using namespace Efont;
 #define INCLUDE_ALTERNATES_OPT	325
 #define EXCLUDE_ALTERNATES_OPT	326
 #define ALTSELECTOR_FEATURE_OPT	327
+#define DEFAULT_LIGKERN_OPT	328
+#define NO_ECOMMAND_OPT		329
+#define LETTER_FEATURE_OPT	330
 
-#define AUTOMATIC_OPT		331
-#define FONT_NAME_OPT		332
-#define QUIET_OPT		333
-#define GLYPHLIST_OPT		334
-#define VENDOR_OPT		335
-#define TYPEFACE_OPT		336
-#define NOCREATE_OPT		337
-#define VERBOSE_OPT		338
+#define AUTOMATIC_OPT		341
+#define FONT_NAME_OPT		342
+#define QUIET_OPT		343
+#define GLYPHLIST_OPT		344
+#define VENDOR_OPT		345
+#define TYPEFACE_OPT		346
+#define NOCREATE_OPT		347
+#define VERBOSE_OPT		348
+#define FORCE_OPT		349
 
-#define VIRTUAL_OPT		341
-#define PL_OPT			342
-#define TFM_OPT			343
-#define MAP_FILE_OPT		344
+#define VIRTUAL_OPT		351
+#define PL_OPT			352
+#define TFM_OPT			353
+#define MAP_FILE_OPT		354
 
-#define DIR_OPTS		350
+#define DIR_OPTS		360
 #define ENCODING_DIR_OPT	(DIR_OPTS + O_ENCODING)
 #define TFM_DIR_OPT		(DIR_OPTS + O_TFM)
 #define PL_DIR_OPT		(DIR_OPTS + O_PL)
@@ -102,7 +107,7 @@ using namespace Efont;
 #define VPL_DIR_OPT		(DIR_OPTS + O_VPL)
 #define TYPE1_DIR_OPT		(DIR_OPTS + O_TYPE1)
 
-#define NO_OUTPUT_OPTS		370
+#define NO_OUTPUT_OPTS		380
 #define NO_ENCODING_OPT		(NO_OUTPUT_OPTS + G_ENCODING)
 #define NO_TYPE1_OPT		(NO_OUTPUT_OPTS + G_TYPE1)
 #define NO_DOTLESSJ_OPT		(NO_OUTPUT_OPTS + G_DOTLESSJ)
@@ -113,6 +118,8 @@ static Clp_Option options[] = {
     
     { "script", 's', SCRIPT_OPT, Clp_ArgString, 0 },
     { "feature", 'f', FEATURE_OPT, Clp_ArgString, 0 },
+    { "letter-feature", 0, LETTER_FEATURE_OPT, Clp_ArgString, 0 },
+    { "lf", 0, LETTER_FEATURE_OPT, Clp_ArgString, 0 },
     { "encoding", 'e', ENCODING_OPT, Clp_ArgString, 0 },
     { "literal-encoding", 0, LITERAL_ENCODING_OPT, Clp_ArgString, 0 },
     { "extend", 'E', EXTEND_OPT, Clp_ArgDouble, 0 },
@@ -123,6 +130,8 @@ static Clp_Option options[] = {
     { "minimum-kern", 'k', MINIMUM_KERN_OPT, Clp_ArgDouble, 0 },
     { "kern-precision", 'k', MINIMUM_KERN_OPT, Clp_ArgDouble, 0 },
     { "ligkern", 0, LIGKERN_OPT, Clp_ArgString, 0 },
+    { "no-encoding-commands", 0, NO_ECOMMAND_OPT, 0, Clp_OnlyNegated },
+    { "default-ligkern", 0, DEFAULT_LIGKERN_OPT, 0, 0 },
     { "unicoding", 0, UNICODING_OPT, Clp_ArgString, 0 },
     { "coding-scheme", 0, CODINGSCHEME_OPT, Clp_ArgString, 0 },
     { "boundary-char", 0, BOUNDARY_CHAR_OPT, CHAR_OPTTYPE, 0 },
@@ -155,6 +164,7 @@ static Clp_Option options[] = {
     { "quiet", 'q', QUIET_OPT, 0, Clp_Negate },
     { "glyphlist", 0, GLYPHLIST_OPT, Clp_ArgString, 0 },
     { "no-create", 0, NOCREATE_OPT, 0, Clp_OnlyNegated },
+    { "force", 0, FORCE_OPT, 0, Clp_Negate },
     { "verbose", 'V', VERBOSE_OPT, 0, Clp_Negate },
     { "kpathsea-debug", 0, KPATHSEA_DEBUG_OPT, Clp_ArgInt, 0 },
 
@@ -169,15 +179,13 @@ static Clp_Option options[] = {
     
 };
 
-static const char * const default_ligkerns[] = {
-   "space l =: lslash ; space L =: Lslash ;",
-   "question quoteleft =: questiondown ;",
-   "exclam quoteleft =: exclamdown ;",
-   "hyphen hyphen =: endash ; endash hyphen =: emdash ;",
-   "quoteleft quoteleft =: quotedblleft ;",
-   "quoteright quoteright =: quotedblright ;",
-   0
-};
+static const char * const default_ligkerns = "\
+space l =: lslash ; space L =: Lslash ; \
+question quoteleft =: questiondown ; \
+exclam quoteleft =: exclamdown ; \
+hyphen hyphen =: endash ; endash hyphen =: emdash ; \
+quoteleft quoteleft =: quotedblleft ; \
+quoteright quoteright =: quotedblright ;";
 
 static const char *program_name;
 static String::Initializer initializer;
@@ -190,9 +198,10 @@ static PermString dot_notdef(".notdef");
 static Vector<Efont::OpenType::Tag> interesting_scripts;
 static Vector<Efont::OpenType::Tag> interesting_features;
 static Vector<Efont::OpenType::Tag> altselector_features;
+static GlyphFilter null_filter;
+static HashMap<Efont::OpenType::Tag, GlyphFilter*> feature_filters(&null_filter);
 
-static Vector<String> include_alternates;
-static Vector<String> exclude_alternates;
+static GlyphFilter alternate_filter;
 
 static String font_name;
 static String encoding_file;
@@ -211,6 +220,7 @@ bool automatic = false;
 bool verbose = false;
 bool no_create = false;
 bool quiet = false;
+bool force = false;
 
 
 void
@@ -243,6 +253,7 @@ Usage: %s [-a] [OPTIONS] OTFFILE FONTNAME\n\n",
 Font feature and transformation options:\n\
   -s, --script=SCRIPT[.LANG]   Use features for script SCRIPT[.LANG] [latn].\n\
   -f, --feature=FEAT           Activate feature FEAT.\n\
+  --lf, --letter-feature=FEAT  Activate feature FEAT for letters.\n\
   -E, --extend=F               Widen characters by a factor of F.\n\
   -S, --slant=AMT              Oblique characters by AMT, generally <<1.\n\
   -L, --letterspacing=AMT      Letterspace each character by AMT units.\n\
@@ -254,6 +265,8 @@ Encoding options:\n\
       --literal-encoding=FILE  Use DVIPS encoding FILE verbatim.\n\
       --ligkern=COMMAND        Add a LIGKERN command.\n\
       --unicoding=COMMAND      Add a UNICODING command.\n\
+      --no-encoding-commands   Ignore encoding file's LIGKERN/UNICODINGs.\n\
+      --default-ligkern        Ignore encoding file's LIGKERNs, use defaults.\n\
       --coding-scheme=SCHEME   Set the output coding scheme to SCHEME.\n\
       --boundary-char=CHAR     Set the boundary character to CHAR.\n\
       --altselector-char=CHAR  Set the alternate selector character to CHAR.\n\
@@ -289,7 +302,8 @@ File location options:\n\
 Other options:\n\
       --glyphlist=FILE         Use FILE to map Adobe glyph names to Unicode.\n\
   -V, --verbose                Print progress information to standard error.\n\
-      --no-create              Print messages, don't modify any files.\n"
+      --no-create              Print messages, don't modify any files.\n\
+      --force                  Generate files even if versions already exist.\n"
 #if HAVE_KPATHSEA
 "      --kpathsea-debug=MASK    Set path searching debug flags to MASK.\n"
 #endif
@@ -590,7 +604,12 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 		Setting &s = settings[j];
 		if (s.op == Setting::SHOW) {
 		    boundser.char_bounds(program->glyph_context(s.y));
-		    sa << "      (SETCHAR " << glyph_ids[s.x] << ')' << glyph_base_comments[s.x] << "\n";
+		    // 3.Aug.2004 -- reported by Marco Kuhlmann: Don't use
+		    // glyph_ids[] array when looking at a different font.
+		    if (program == family_cff)
+			sa << "      (SETCHAR " << glyph_ids[s.x] << ')' << glyph_base_comments[s.x] << "\n";
+		    else
+			sa << "      (SETCHAR D " << s.x << ")\n";
 		} else if (s.op == Setting::MOVE && vpl) {
 		    boundser.translate(s.x, s.y);
 		    if (s.x)
@@ -643,16 +662,19 @@ output_pl(const Metrics &metrics, const String &ps_name, int boundary_char,
 
 struct Lookup {
     bool used;
+    bool required;
     Vector<OpenType::Tag> features;
-    Lookup()			: used(false) { }
+    GlyphFilter* filter;
+    Lookup()			: used(false), required(false), filter(0) { }
 };
 
 static void
-find_lookups(const OpenType::ScriptList &scripts, const OpenType::FeatureList &features, Vector<Lookup> &lookups, ErrorHandler *errh)
+find_lookups(const OpenType::ScriptList& scripts, const OpenType::FeatureList& features, Vector<Lookup>& lookups, ErrorHandler* errh)
 {
     Vector<int> fids, lookupids;
     int required;
-    
+
+    // go over all scripts
     for (int i = 0; i < interesting_scripts.size(); i += 2) {
 	OpenType::Tag script = interesting_scripts[i];
 	OpenType::Tag langsys = interesting_scripts[i+1];
@@ -676,10 +698,23 @@ find_lookups(const OpenType::ScriptList &scripts, const OpenType::FeatureList &f
 		else {
 		    lookups[l].used = true;
 		    lookups[l].features.push_back(ftag);
+		    if (j < 0)
+			lookups[l].required = true;
 		}
 	    }
 	}
     }
+
+    // now check for compatible glyph filters
+    for (Lookup* l = lookups.begin(); l < lookups.end(); l++)
+	if (l->used && !l->required) {
+	    l->filter = feature_filters[l->features[0]];
+	    for (OpenType::Tag* ft = l->features.begin() + 1; ft < l->features.end(); ft++)
+		if (!l->filter->check_eq(*feature_filters[*ft])) {
+		    errh->error("'%s' and '%s' features share a lookup, but have different filters", l->features[0].text().c_str(), ft->text().c_str());
+		    break;
+		}
+	}
 }
 
 static int
@@ -1015,12 +1050,12 @@ enum { F_GSUB_TRY = 1, F_GSUB_PART = 2, F_GSUB_ALL = 4,
        X_GPOS_NONE = 4, X_GPOS_PART = 5, X_COUNT };
 
 static const char * const x_messages[] = {
-    "this script does not support '%s'",
-    "'%s' ignored, too complex for me",
-    "complex substitutions from '%s' ignored",
-    "some complex substitutions from '%s' ignored",
-    "complex positionings from '%s' ignored",
-    "some complex positionings from '%s' ignored",
+    "% ignored, not supported by font",
+    "% ignored, too complex for me",
+    "complex substitutions from % ignored",
+    "some complex substitutions from % ignored",
+    "complex positionings from % ignored",
+    "some complex positionings from % ignored",
 };
 
 static void
@@ -1030,22 +1065,23 @@ report_underused_features(const HashMap<uint32_t, int> &feature_usage, ErrorHand
     for (int i = 0; i < interesting_features.size(); i++) {
 	OpenType::Tag f = interesting_features[i];
 	int fu = feature_usage[f.value()];
+	String ftext = "'" + f.text() + "'";
 	if (fu == 0)
-	    x[X_UNUSED].push_back(f.text());
+	    x[X_UNUSED].push_back(ftext);
 	else if ((fu & (F_GSUB_TRY | F_GPOS_TRY)) == fu)
-	    x[X_BOTH_NONE].push_back(f.text());
+	    x[X_BOTH_NONE].push_back(ftext);
 	else {
 	    if (fu & F_GSUB_TRY) {
 		if ((fu & (F_GSUB_PART | F_GSUB_ALL)) == 0)
-		    x[X_GSUB_NONE].push_back(f.text());
+		    x[X_GSUB_NONE].push_back(ftext);
 		else if (fu & F_GSUB_PART)
-		    x[X_GSUB_PART].push_back(f.text());
+		    x[X_GSUB_PART].push_back(ftext);
 	    }
 	    if (fu & F_GPOS_TRY) {
 		if ((fu & (F_GPOS_PART | F_GPOS_ALL)) == 0)
-		    x[X_GPOS_NONE].push_back(f.text());
+		    x[X_GPOS_NONE].push_back(ftext);
 		else if (fu & F_GPOS_PART)
-		    x[X_GPOS_PART].push_back(f.text());
+		    x[X_GPOS_PART].push_back(ftext);
 	    }
 	}
     }
@@ -1059,9 +1095,15 @@ report_underused_features(const HashMap<uint32_t, int> &feature_usage, ErrorHand
     for (int i = 0; i < X_COUNT; i++)
 	if (x[i].size()) {
 	    StringAccum sa;
-	    sa.append_fill_lines(x[i], 65, "", "", ", ");
-	    sa.pop_back();
-	    errh->warning(x_messages[i], sa.c_str());
+	    const char* msg_pct = strchr(x_messages[i], '%');
+	    sa.append(x_messages[i], msg_pct - x_messages[i]);
+	    const char* sep = (x[i].size() > 2 ? ", " : " ");
+	    for (const String* a = x[i].begin(); a < x[i].end() - 1; a++)
+		sa << *a << sep;
+	    sa << (x[i].size() > 1 ? "and " : "") << x[i].back()
+	       << (x[i].size() > 1 ? " features" : " feature") << (msg_pct+1);
+	    sa.append_break_lines(sa.take_string(), 61);
+	    errh->warning("%s", sa.c_str());
 	}
 }
 
@@ -1076,6 +1118,92 @@ main_dvips_map(const String &ps_name, StringAccum &sa, ErrorHandler *errh)
 	sa << "<<" << pathname_filename(otf_filename);
 }
 
+static void
+do_gsub(Metrics& metrics, const OpenType::Font& otf, DvipsEncoding& dvipsenc, bool dvipsenc_literal, HashMap<uint32_t, int>& feature_usage, const Vector<PermString>& glyph_names, ErrorHandler* errh)
+{
+    // apply activated GSUB features
+    OpenType::Gsub gsub(otf.table("GSUB"), errh);
+    Vector<Lookup> lookups(gsub.nlookups(), Lookup());
+    find_lookups(gsub.script_list(), gsub.feature_list(), lookups, errh);
+    Vector<OpenType::Substitution> subs;
+    for (int i = 0; i < lookups.size(); i++)
+	if (lookups[i].used) {
+	    OpenType::GsubLookup l = gsub.lookup(i);
+	    subs.clear();
+	    bool understood = l.unparse_automatics(gsub, subs);
+
+	    // check for -ffina, which should apply only at the ends of words,
+	    // and -finit, which should apply only at the beginnings.
+	    OpenType::Tag feature = (lookups[i].features.size() == 1 ? lookups[i].features[0] : OpenType::Tag());
+	    if (feature == OpenType::Tag("fina") || feature == OpenType::Tag("fin2") || feature == OpenType::Tag("fin3")) {
+		if (dvipsenc.boundary_char() < 0)
+		    errh->warning("'-ffina' requires a boundary character\n(The input encoding didn't specify a boundary character, but\nI need one to implement '-ffina' features correctly.  Try\nthe '--boundary-char' option.)");
+		else {
+		    int bg = metrics.boundary_glyph();
+		    for (int j = 0; j < subs.size(); j++)
+			subs[j].add_outer_right(bg);
+		}
+	    } else if (feature == OpenType::Tag("init")) {
+		int bg = metrics.boundary_glyph();
+		for (int j = 0; j < subs.size(); j++)
+		    subs[j].add_outer_left(bg);
+	    }
+
+	    //for (int subno = 0; subno < subs.size(); subno++) fprintf(stderr, "%5d\t%s\n", i, subs[subno].unparse().c_str());
+
+	    // figure out which glyph filter to use
+	    int nunderstood = metrics.apply(subs, !dvipsenc_literal, i, *lookups[i].filter + alternate_filter, glyph_names);
+
+	    // mark as used
+	    int d = (understood && nunderstood == subs.size() ? F_GSUB_ALL : (nunderstood ? F_GSUB_PART : 0)) + F_GSUB_TRY;
+	    for (int j = 0; j < lookups[i].features.size(); j++)
+		feature_usage.find_force(lookups[i].features[j].value()) |= d;
+	}
+
+    // apply 'aalt' feature if we have variant selectors
+    if (metrics.altselectors() && !dvipsenc_literal) {
+	// apply default features
+	if (!altselector_features.size()) {
+	    altselector_features.push_back(OpenType::Tag("dlig"));
+	    altselector_features.push_back(OpenType::Tag("salt"));
+	}
+	// do lookups
+	altselector_features.swap(interesting_features);
+	Vector<Lookup> alt_lookups(gsub.nlookups(), Lookup());
+	find_lookups(gsub.script_list(), gsub.feature_list(), alt_lookups, ErrorHandler::silent_handler());
+	Vector<OpenType::Substitution> alt_subs;
+	for (int i = 0; i < alt_lookups.size(); i++)
+	    if (alt_lookups[i].used) {
+		OpenType::GsubLookup l = gsub.lookup(i);
+		alt_subs.clear();
+		(void) l.unparse_automatics(gsub, alt_subs);
+		metrics.apply_alternates(alt_subs, i, alternate_filter, glyph_names);
+	    }
+	altselector_features.swap(interesting_features);
+    }
+}
+
+static void
+do_gpos(Metrics& metrics, const OpenType::Font& otf, HashMap<uint32_t, int>& feature_usage, ErrorHandler* errh)
+{
+    OpenType::Gpos gpos(otf.table("GPOS"), errh);
+    Vector<Lookup> lookups(gpos.nlookups(), Lookup());
+    find_lookups(gpos.script_list(), gpos.feature_list(), lookups, errh);
+    Vector<OpenType::Positioning> poss;
+    for (int i = 0; i < lookups.size(); i++)
+	if (lookups[i].used) {
+	    OpenType::GposLookup l = gpos.lookup(i);
+	    poss.clear();
+	    bool understood = l.unparse_automatics(poss);
+	    int nunderstood = metrics.apply(poss);
+
+	    // mark as used
+	    int d = (understood && nunderstood == poss.size() ? F_GPOS_ALL : (nunderstood ? F_GPOS_PART : 0)) + F_GPOS_TRY;
+	    for (int j = 0; j < lookups[i].features.size(); j++)
+		feature_usage.find_force(lookups[i].features[j].value()) |= d;
+	}
+}
+
 static void
 do_file(const String &otf_filename, const OpenType::Font &otf,
 	const DvipsEncoding &dvipsenc_in, bool dvipsenc_literal,
@@ -1119,76 +1247,25 @@ do_file(const String &otf_filename, const OpenType::Font &otf,
     OpenType::Cmap cmap(otf.table("cmap"), errh);
     assert(cmap.ok());
     if (dvipsenc_literal)
-	dvipsenc.make_literal_metrics(metrics, font);
+	dvipsenc.make_metrics(metrics, cmap, font, 0, true, errh);
     else {
 	T1Secondary secondary(font, cmap);
 	secondary.set_font_information(font_name, otf, otf_filename);
-	dvipsenc.make_metrics(metrics, cmap, font, &secondary, errh);
+	dvipsenc.make_metrics(metrics, cmap, font, &secondary, false, errh);
     }
-    // encode boundary glyph at 256
-    metrics.encode(256, metrics.boundary_glyph());
-    
+    // encode boundary glyph at 256; pretend its Unicode value is '\n'
+    metrics.encode(256, '\n', metrics.boundary_glyph());
+
     // maintain statistics about features
     HashMap<uint32_t, int> feature_usage(0);
-    
+
     // apply activated GSUB features
-    OpenType::Gsub gsub(otf.table("GSUB"), errh);
-    Vector<Lookup> lookups(gsub.nlookups(), Lookup());
-    find_lookups(gsub.script_list(), gsub.feature_list(), lookups, errh);
-    Vector<OpenType::Substitution> subs;
-    for (int i = 0; i < lookups.size(); i++)
-	if (lookups[i].used) {
-	    OpenType::GsubLookup l = gsub.lookup(i);
-	    subs.clear();
-	    bool understood = l.unparse_automatics(gsub, subs);
-
-	    // check for -ffina, which should apply only at the ends of words,
-	    // and -finit, which should apply only at the beginnings.
-	    OpenType::Tag feature = (lookups[i].features.size() == 1 ? lookups[i].features[0] : OpenType::Tag());
-	    if (feature == OpenType::Tag("fina") || feature == OpenType::Tag("fin2") || feature == OpenType::Tag("fin3")) {
-		if (dvipsenc.boundary_char() < 0)
-		    errh->warning("'-ffina' requires a boundary character\n(The input encoding didn't specify a boundary character, but\nI need one to implement '-ffina' features correctly. Add one\nwith a \"%% LIGKERN || = <slot> ;\" command in the encoding.)");
-		else {
-		    int bg = metrics.boundary_glyph();
-		    for (int j = 0; j < subs.size(); j++)
-			subs[j].add_outer_right(bg);
-		}
-	    } else if (feature == OpenType::Tag("init")) {
-		int bg = metrics.boundary_glyph();
-		for (int j = 0; j < subs.size(); j++)
-		    subs[j].add_outer_left(bg);
-	    }
-
-	    //for (int subno = 0; subno < subs.size(); subno++) fprintf(stderr, "%5d\t%s\n", i, subs[subno].unparse().c_str());
-	    
-	    int nunderstood = metrics.apply(subs, !dvipsenc_literal, i, include_alternates, exclude_alternates, glyph_names);
-
-	    // mark as used
-	    int d = (understood && nunderstood == subs.size() ? F_GSUB_ALL : (nunderstood ? F_GSUB_PART : 0)) + F_GSUB_TRY;
-	    for (int j = 0; j < lookups[i].features.size(); j++)
-		feature_usage.find_force(lookups[i].features[j].value()) |= d;
-	}
-
-    // apply 'aalt' feature if we have variant selectors
-    if (metrics.altselectors() && !dvipsenc_literal) {
-	// apply default features
-	if (!altselector_features.size()) {
-	    altselector_features.push_back(OpenType::Tag("dlig"));
-	    altselector_features.push_back(OpenType::Tag("salt"));
-	}
-	// do lookups
-	altselector_features.swap(interesting_features);
-	Vector<Lookup> alt_lookups(gsub.nlookups(), Lookup());
-	find_lookups(gsub.script_list(), gsub.feature_list(), alt_lookups, ErrorHandler::silent_handler());
-	Vector<OpenType::Substitution> alt_subs;
-	for (int i = 0; i < alt_lookups.size(); i++)
-	    if (alt_lookups[i].used) {
-		OpenType::GsubLookup l = gsub.lookup(i);
-		alt_subs.clear();
-		(void) l.unparse_automatics(gsub, alt_subs);
-		metrics.apply_alternates(alt_subs, i, include_alternates, exclude_alternates, glyph_names);
-	    }
-	altselector_features.swap(interesting_features);
+    try {
+	do_gsub(metrics, otf, dvipsenc, dvipsenc_literal, feature_usage, glyph_names, errh);
+    } catch (OpenType::BlankTable) {
+	// nada
+    } catch (OpenType::Error e) {
+	errh->warning("GSUB '%s' error, continuing", e.description.c_str());
     }
     
     // apply LIGKERN ligature commands to the result
@@ -1202,22 +1279,13 @@ do_file(const String &otf_filename, const OpenType::Font &otf,
 	metrics.shrink_encoding(257, dvipsenc_in, errh);
     
     // apply activated GPOS features
-    OpenType::Gpos gpos(otf.table("GPOS"), errh);
-    lookups.assign(gpos.nlookups(), Lookup());
-    find_lookups(gpos.script_list(), gpos.feature_list(), lookups, errh);
-    Vector<OpenType::Positioning> poss;
-    for (int i = 0; i < lookups.size(); i++)
-	if (lookups[i].used) {
-	    OpenType::GposLookup l = gpos.lookup(i);
-	    poss.clear();
-	    bool understood = l.unparse_automatics(poss);
-	    int nunderstood = metrics.apply(poss);
-
-	    // mark as used
-	    int d = (understood && nunderstood == poss.size() ? F_GPOS_ALL : (nunderstood ? F_GPOS_PART : 0)) + F_GPOS_TRY;
-	    for (int j = 0; j < lookups[i].features.size(); j++)
-		feature_usage.find_force(lookups[i].features[j].value()) |= d;
-	}
+    try {
+	do_gpos(metrics, otf, feature_usage, errh);
+    } catch (OpenType::BlankTable) {
+	// nada
+    } catch (OpenType::Error e) {
+	errh->warning("GPOS '%s' error, continuing", e.description.c_str());
+    }
 
     // apply LIGKERN kerning commands to the result
     dvipsenc.apply_ligkern_kern(metrics, errh);
@@ -1343,8 +1411,9 @@ main(int argc, char *argv[])
     bool literal_encoding = false;
     Vector<String> ligkern;
     Vector<String> unicoding;
+    bool no_ecommand = false, default_ligkern = false;
     String codingscheme;
-  
+    
     while (1) {
 	int opt = Clp_Next(clp);
 	switch (opt) {
@@ -1377,6 +1446,18 @@ main(int argc, char *argv[])
 	      break;
 	  }
       
+	  case LETTER_FEATURE_OPT: {
+	      OpenType::Tag t(clp->arg);
+	      if (t.valid()) {
+		  interesting_features.push_back(t);
+		  GlyphFilter* gf = new GlyphFilter;
+		  gf->add_substitution_filter("<Letter>", false, errh);
+		  feature_filters.insert(t, gf); 
+	      } else
+		  usage_error(errh, "bad feature tag");
+	      break;
+	  }
+
 	  case ENCODING_OPT:
 	    if (encoding_file)
 		usage_error(errh, "encoding specified twice");
@@ -1420,6 +1501,14 @@ main(int argc, char *argv[])
 	    ligkern.push_back(clp->arg);
 	    break;
 
+	  case NO_ECOMMAND_OPT:
+	    no_ecommand = true;
+	    break;
+
+	  case DEFAULT_LIGKERN_OPT:
+	    default_ligkern = true;
+	    break;
+
 	  case BOUNDARY_CHAR_OPT:
 	    ligkern.push_back(String("|| = ") + String(clp->val.i));
 	    break;
@@ -1439,14 +1528,15 @@ main(int argc, char *argv[])
 
 	  case EXCLUDE_ALTERNATES_OPT:
 	  case INCLUDE_ALTERNATES_OPT: {
-	      Vector<String> *v = (opt == EXCLUDE_ALTERNATES_OPT ? &exclude_alternates : &include_alternates);
 	      const char *s = clp->arg;
 	      while (*s) {
 		  const char *start = s;
 		  while (*s && !isspace(*s))
 		      s++;
-		  if (s > start)
-		      v->push_back(String(start, s - start));
+		  if (s > start) {
+		      String str(start, s - start);
+		      alternate_filter.add_alternate_filter(str, opt == EXCLUDE_ALTERNATES_OPT, errh);
+		  }
 		  while (isspace(*s))
 		      s++;
 	      }
@@ -1560,6 +1650,10 @@ main(int argc, char *argv[])
 	    no_create = clp->negated;
 	    break;
 
+	  case FORCE_OPT:
+	    force = !clp->negated;
+	    break;
+
 	  case KPATHSEA_DEBUG_OPT:
 #if HAVE_KPATHSEA
 	    kpsei_set_debug_flags(clp->val.u);
@@ -1605,73 +1699,76 @@ particular purpose.\n");
     }
     
   done:
-    if (!input_file)
-	usage_error(errh, "no font filename provided");
-    if (!encoding_file) {
-	errh->warning("no encoding provided");
-	errh->message("(Use '-e ENCODING' to choose an encoding. '-e texnansx' often works,\nor say '-e -' to turn off this warning.)");
-    } else if (encoding_file == "-")
-	encoding_file = "";
+    try {
+	if (!input_file)
+	    usage_error(errh, "no font filename provided");
+	if (!encoding_file) {
+	    errh->warning("no encoding provided");
+	    errh->message("(Use '-e ENCODING' to choose an encoding. '-e texnansx' often works,\nor say '-e -' to turn off this warning.)");
+	} else if (encoding_file == "-")
+	    encoding_file = "";
     
-    // read font
-    String font_data = read_file(input_file, errh);
-    if (errh->nerrors())
-	exit(1);
+	// read font
+	String font_data = read_file(input_file, errh);
+	if (errh->nerrors())
+	    exit(1);
 
-    LandmarkErrorHandler cerrh(errh, printable_filename(input_file));
-    BailErrorHandler bail_errh(&cerrh);
+	LandmarkErrorHandler cerrh(errh, printable_filename(input_file));
+	BailErrorHandler bail_errh(&cerrh);
 
-    OpenType::Font otf(font_data, &bail_errh);
-    assert(otf.ok());
+	OpenType::Font otf(font_data, &bail_errh);
+	assert(otf.ok());
 
-    // figure out scripts we care about
-    if (!interesting_scripts.size()) {
-	interesting_scripts.push_back(Efont::OpenType::Tag("latn"));
-	interesting_scripts.push_back(Efont::OpenType::Tag());
-    }
-    if (interesting_features.size())
+	// figure out scripts we care about
+	if (!interesting_scripts.size()) {
+	    interesting_scripts.push_back(Efont::OpenType::Tag("latn"));
+	    interesting_scripts.push_back(Efont::OpenType::Tag());
+	}
 	std::sort(interesting_features.begin(), interesting_features.end());
 
-    // read glyphlist
-    if (String s = read_file(glyphlist_file, errh, true))
-	DvipsEncoding::parse_glyphlist(s);
+	// read glyphlist
+	if (String s = read_file(glyphlist_file, errh, true))
+	    DvipsEncoding::parse_glyphlist(s);
 
-    // read encoding
-    DvipsEncoding dvipsenc;
-    if (encoding_file) {
-	if (String path = locate_encoding(encoding_file, errh))
-	    dvipsenc.parse(path, errh);
-	else
-	    errh->fatal("encoding '%s' not found", encoding_file.c_str());
-    } else {
-	// use encoding from font
-	Cff cff(otf.table("CFF"), &bail_errh);
-	Cff::FontParent *font = cff.font(PermString(), &bail_errh);
-	assert(cff.ok() && font->ok());
-	if (Type1Encoding *t1e = font->type1_encoding()) {
-	    for (int i = 0; i < 256; i++)
-		dvipsenc.encode(i, (*t1e)[i]);
-	} else
-	    errh->fatal("font has no encoding, specify one explicitly");
-	delete font;
-    }
+	// read encoding
+	DvipsEncoding dvipsenc;
+	if (encoding_file) {
+	    if (String path = locate_encoding(encoding_file, errh))
+		dvipsenc.parse(path, no_ecommand || default_ligkern, no_ecommand, errh);
+	    else
+		errh->fatal("encoding '%s' not found", encoding_file.c_str());
+	} else {
+	    // use encoding from font
+	    Cff cff(otf.table("CFF"), &bail_errh);
+	    Cff::FontParent *font = cff.font(PermString(), &bail_errh);
+	    assert(cff.ok() && font->ok());
+	    if (Type1Encoding *t1e = font->type1_encoding()) {
+		for (int i = 0; i < 256; i++)
+		    dvipsenc.encode(i, (*t1e)[i]);
+	    } else
+		errh->fatal("font has no encoding, specify one explicitly");
+	    delete font;
+	}
 
-    // apply default ligkern commands
-    if (!dvipsenc.file_had_comments() && !ligkern.size())
-	for (const char * const *lk = default_ligkerns; *lk; lk++)
-	    dvipsenc.parse_ligkern(*lk, ErrorHandler::silent_handler());
+	// apply default ligkern commands
+	if (default_ligkern || (!dvipsenc.file_had_ligkern() && !ligkern.size() && !no_ecommand))
+	    dvipsenc.parse_ligkern(default_ligkerns, ErrorHandler::silent_handler());
     
-    // apply command-line ligkern commands and coding scheme
-    cerrh.set_landmark("--ligkern command");
-    for (int i = 0; i < ligkern.size(); i++)
-	dvipsenc.parse_ligkern(ligkern[i], &cerrh);
-    cerrh.set_landmark("--unicoding command");
-    for (int i = 0; i < unicoding.size(); i++)
-	dvipsenc.parse_unicoding(unicoding[i], &cerrh);
-    if (codingscheme)
-	dvipsenc.set_coding_scheme(codingscheme);
+	// apply command-line ligkern commands and coding scheme
+	cerrh.set_landmark("--ligkern command");
+	for (int i = 0; i < ligkern.size(); i++)
+	    dvipsenc.parse_ligkern(ligkern[i], &cerrh);
+	cerrh.set_landmark("--unicoding command");
+	for (int i = 0; i < unicoding.size(); i++)
+	    dvipsenc.parse_unicoding(unicoding[i], &cerrh);
+	if (codingscheme)
+	    dvipsenc.set_coding_scheme(codingscheme);
 
-    do_file(input_file, otf, dvipsenc, literal_encoding, errh);
+	do_file(input_file, otf, dvipsenc, literal_encoding, errh);
+	
+    } catch (OpenType::Error e) {
+	errh->error("unhandled exception '%s'", e.description.c_str());
+    }
     
     return (errh->nerrors() == 0 ? 0 : 1);
 }
